/*
Copyright 2009 Marcio Bueno Licensed under the
Educational Community License, Version 2.0 (the "License"); you may
not use this file except in compliance with the License. You may
obtain a copy of the License at

http://www.osedu.org/licenses/ECL-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an "AS IS"
BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
or implied. See the License for the specific language governing
permissions and limitations under the License.
*/

#ifndef __FILA_H__
#define __FILA_H__

#include <iostream>

using std::ostream;

// Template para poder definir um tipo genrico 'T'
template <typename T>
class Fila
{
private:
	// Estrutura interna privada que representa um n de uma
	// fila simplesmente encadeada
	struct No
	{
		T valor;  // Valor armazenado na fila de um tipo genrico 'T'
		No *prox; // Ponteiro para o prximo elemento

		// Construtor - Inicializa o N com um valor recebido
		No(const T &valor)
		{
			this->valor = valor; // this  utilizado para referenciar o campo do struct
			this->prox = NULL;
		}
	};

	No *inicio, *fim; // Campos privados da Fila (seu incio e seu fim)

public:
	// Construtor que cria uma fila vazia
	Fila() 
	{
		this->inicio = this->fim = NULL;
	}

	// retorna TRUE se for vazia e FALSE caso contrrio
	bool vazia()
	{
		return this->inicio == NULL;
	}

	// retorna TRUE se o valor estiver na fila e FALSE caso contrrio
	bool consultar(const T &valor) const;
	
	// retorna o primeiro valor da fila
	// Cuidado que dar erro ao cham-lo para uma fila vazia
	// Duas verses: uma "const" e outra no
	const T primeiro() const
	{
		return this->inicio->valor;
	}

	T primeiro()
	{
		return this->inicio->valor;
	}

	// retorna o ltimo valor da fila
	// Cuidado que dar erro ao cham-lo para uma fila vazia 
	// Duas verses: uma "const" e outra no
	const T ultimo() const
	{
		return this->fim->valor;
	}
	
	T ultimo()
	{
		return this->fim->valor;
	}

	// insere um valor no final da fila
	void inserir(const T &valor) 
	{
		No *aux = new No(valor);
		if (this->inicio == NULL) //  fila vazia?
		{
			this->inicio = this->fim = aux;
		} 
		else 
		{
			this->fim->prox = aux;
			this->fim = aux;
		}
	}

	// operador += para funcionar como o inserir em lista
	Fila<T> &operator += (const T &valor)
	{
		this->inserir(valor);
		return *this;
	}

	// remover o valor do incio da fila e devolv-lo como retorno da funo
	// Cuidado que dar erro ao cham-lo para uma fila vazia 
	T remover() 
	{
		T aux = this->inicio->valor;
		this->inicio = this->inicio->prox;

		// verifica se a lista ficou vazia
		if (this->inicio == NULL)
		{
			this->fim = NULL;
		}

		return aux;
	}

	// Imprime uma fila na sada padro no formato: {} ou {a, b, c, ...}
	template<typename T>
	friend ostream &operator << (ostream &out, Fila<T> &f);
};

// retorna TRUE se o valor estiver na fila e FALSE caso contrrio
template <typename T>
bool Fila<T>::consultar(const T &valor) const
{
	No *aux = this->inicio;
	while (aux != NULL)
	{
		if (aux->valor == valor)
			return true;
		aux = aux->prox;
	}
	return false;
}

// Imprime uma fila na sada padro no formato: {} ou {a, b, c, ...}
template <typename T>
ostream &operator << (ostream &out, Fila<T> &f) {
	out << '{';
	if (f.inicio != NULL) {
		Fila<T>::No *aux = f.inicio->prox;
		out << f.inicio->valor;
		while (aux != NULL) {
			out << ", " << aux->valor;
			aux = aux->prox;
		}
	}
	out << '}';
	return out;
}

#endif // __FILA_H__